@@ -8,7 +8,7 @@ var (
 	// MaxStackVarSize is the maximum size variable which we will allocate on the stack.
 	// This limit is for explicit variable declarations like "var x T" or "x := ...".
 	// Note: the flag smallframes can update this value.
-	MaxStackVarSize = int64(10 * 1024 * 1024)
+	MaxStackVarSize = int64(128 * 1024)
 
 	// MaxImplicitStackVarSize is the maximum size of implicit variables that we will allocate on the stack.
 	//   p := new(T)          allocating T on the stack
