@@ -7,6 +7,7 @@ package ssa_test
 import (
 	cmddwarf "cmd/internal/dwarf"
 	"cmd/internal/quoted"
+	"cmp"
 	"debug/dwarf"
 	"debug/elf"
 	"debug/macho"
@@ -18,7 +19,8 @@ import (
 	"io"
 	"os"
 	"runtime"
-	"sort"
+	"slices"
+	"strings"
 	"testing"
 )
 
@@ -101,7 +103,7 @@ func TestStmtLines(t *testing.T) {
 		if pkgname == "runtime" {
 			continue
 		}
-		if pkgname == "crypto/internal/nistec/fiat" {
+		if pkgname == "crypto/internal/fips140/nistec/fiat" {
 			continue // golang.org/issue/49372
 		}
 		if e.Val(dwarf.AttrStmtList) == nil {
@@ -118,6 +120,11 @@ func TestStmtLines(t *testing.T) {
 				break
 			}
 			must(err)
+			if le.EndSequence {
+				// When EndSequence is true only
+				// le.Address is meaningful, skip.
+				continue
+			}
 			fl := Line{le.File.Name, le.Line}
 			lines[fl] = lines[fl] || le.IsStmt
 		}
@@ -144,11 +151,11 @@ func TestStmtLines(t *testing.T) {
 	}
 	t.Logf("Saw %d out of %d lines without statement marks", len(nonStmtLines), len(lines))
 	if testing.Verbose() {
-		sort.Slice(nonStmtLines, func(i, j int) bool {
-			if nonStmtLines[i].File != nonStmtLines[j].File {
-				return nonStmtLines[i].File < nonStmtLines[j].File
+		slices.SortFunc(nonStmtLines, func(a, b Line) int {
+			if a.File != b.File {
+				return strings.Compare(a.File, b.File)
 			}
-			return nonStmtLines[i].Line < nonStmtLines[j].Line
+			return cmp.Compare(a.Line, b.Line)
 		})
 		for _, l := range nonStmtLines {
 			t.Logf("%s:%d has no DWARF is_stmt mark\n", l.File, l.Line)
