@@ -17,9 +17,11 @@ information on how to use it see the cgo documentation (go doc cmd/cgo).
 
 The second is the SWIG program, which is a general tool for
 interfacing between languages. For information on SWIG see
-http://swig.org/. When running go build, any file with a .swig
+https://swig.org/. When running go build, any file with a .swig
 extension will be passed to SWIG. Any file with a .swigcxx extension
-will be passed to SWIG with the -c++ option.
+will be passed to SWIG with the -c++ option. A package can't be just
+a .swig or .swigcxx file; there must be at least one .go file, even if
+it has just a package clause.
 
 When either cgo or SWIG is used, go build will pass any .c, .m, .s, .S
 or .sx files to the C compiler, and any .cc, .cpp, .cxx files to the C++
@@ -34,7 +36,7 @@ var HelpPackages = &base.Command{
 	Long: `
 Many commands apply to a set of packages:
 
-	go action [packages]
+	go <action> [packages]
 
 Usually, [packages] is a list of import paths.
 
@@ -49,16 +51,14 @@ environment variable (For more details see: 'go help gopath').
 If no import paths are given, the action applies to the
 package in the current directory.
 
-There are four reserved names for paths that should not be used
+There are five reserved names for paths that should not be used
 for packages to be built with the go tool:
 
 - "main" denotes the top-level package in a stand-alone executable.
 
-- "all" expands to all packages found in all the GOPATH
-trees. For example, 'go list all' lists all the packages on the local
-system. When using modules, "all" expands to all packages in
-the main module and their dependencies, including dependencies
-needed by tests of any of those.
+- "all" expands to all packages in the main module (or workspace modules) and
+their dependencies, including dependencies needed by tests of any of those. In
+GOPATH mode, "all" expands to all packages found in all the GOPATH trees.
 
 - "std" is like all but expands to just the packages in the standard
 Go library.
@@ -66,6 +66,14 @@ Go library.
 - "cmd" expands to the Go repository's commands and their
 internal libraries.
 
+- "tool" expands to the tools defined in the current module's go.mod file.
+
+Package names match against fully-qualified import paths or patterns that
+match against any number of import paths. For instance, "fmt" refers to the
+standard library's package fmt, but "http" alone for package http would not
+match the import path "net/http" from the standard library. Instead, the
+complete import path "net/http" must be used.
+
 Import paths beginning with "cmd/" only match source code in
 the Go repository.
 
@@ -95,7 +103,10 @@ By convention, this is arranged by starting each path with a
 unique prefix that belongs to you. For example, paths used
 internally at Google all begin with 'google', and paths
 denoting remote repositories begin with the path to the code,
-such as 'github.com/user/repo'.
+such as 'github.com/user/repo'. Package patterns should include this prefix.
+For instance, a package called 'http' residing under 'github.com/user/repo',
+would be addressed with the fully-qualified pattern:
+'github.com/user/repo/http'.
 
 Packages in a program need not have unique package names,
 but there are two reserved package names with special meaning.
@@ -205,12 +216,12 @@ For example,
 	import "example.org/user/foo.hg"
 
 denotes the root directory of the Mercurial repository at
-example.org/user/foo or foo.hg, and
+example.org/user/foo, and
 
 	import "example.org/repo.git/foo/bar"
 
 denotes the foo/bar directory of the Git repository at
-example.org/repo or repo.git.
+example.org/repo.
 
 When a version control system supports multiple protocols,
 each is tried in turn when downloading. For example, a Git
@@ -230,6 +241,11 @@ The meta tag has the form:
 
 	<meta name="go-import" content="import-prefix vcs repo-root">
 
+Starting in Go 1.25, an optional subdirectory will be recognized by the
+go command:
+
+	<meta name="go-import" content="import-prefix vcs repo-root subdir">
+
 The import-prefix is the import path corresponding to the repository
 root. It must be a prefix or an exact match of the package being
 fetched with "go get". If it's not an exact match, another http
@@ -244,6 +260,12 @@ The vcs is one of "bzr", "fossil", "git", "hg", "svn".
 The repo-root is the root of the version control system
 containing a scheme and not containing a .vcs qualifier.
 
+The subdir specifies the directory within the repo-root where the
+Go module's root (including its go.mod file) is located. It allows
+you to organize your repository with the Go module code in a subdirectory
+rather than directly at the repository's root.
+If set, all vcs tags must be prefixed with "subdir". i.e. "subdir/v1.2.3"
+
 For example,
 
 	import "example.org/pkg/foo"
@@ -258,14 +280,17 @@ If that page contains the meta tag
 	<meta name="go-import" content="example.org git https://code.org/r/p/exproj">
 
 the go tool will verify that https://example.org/?go-get=1 contains the
-same meta tag and then git clone https://code.org/r/p/exproj into
-GOPATH/src/example.org.
+same meta tag and then download the code from the Git repository at https://code.org/r/p/exproj
 
-When using GOPATH, downloaded packages are written to the first directory
-listed in the GOPATH environment variable.
-(See 'go help gopath-get' and 'go help gopath'.)
+If that page contains the meta tag
 
-When using modules, downloaded packages are stored in the module cache.
+	<meta name="go-import" content="example.org git https://code.org/r/p/exproj foo/subdir">
+
+the go tool will verify that https://example.org/?go-get=1 contains the same meta
+tag and then download the code from the "foo/subdir" subdirectory within the Git repository
+at https://code.org/r/p/exproj
+
+Downloaded packages are stored in the module cache.
 See https://golang.org/ref/mod#module-cache.
 
 When using modules, an additional variant of the go-import meta tag is
@@ -471,9 +496,9 @@ var HelpEnvironment = &base.Command{
 	Long: `
 
 The go command and the tools it invokes consult environment variables
-for configuration. If an environment variable is unset, the go command
-uses a sensible default setting. To see the effective setting of the
-variable <NAME>, run 'go env <NAME>'. To change the default setting,
+for configuration. If an environment variable is unset or empty, the go
+command uses a sensible default setting. To see the effective setting of
+the variable <NAME>, run 'go env <NAME>'. To change the default setting,
 run 'go env -w <NAME>=<VALUE>'. Defaults changed using 'go env -w'
 are recorded in a Go environment configuration file stored in the
 per-user configuration directory, as reported by os.UserConfigDir.
@@ -484,25 +509,31 @@ See 'go help env' for details.
 
 General-purpose environment variables:
 
+	GCCGO
+		The gccgo command to run for 'go build -compiler=gccgo'.
 	GO111MODULE
 		Controls whether the go command runs in module-aware mode or GOPATH mode.
 		May be "off", "on", or "auto".
 		See https://golang.org/ref/mod#mod-commands.
-	GCCGO
-		The gccgo command to run for 'go build -compiler=gccgo'.
 	GOARCH
 		The architecture, or processor, for which to compile code.
 		Examples are amd64, 386, arm, ppc64.
+	GOAUTH
+		Controls authentication for go-import and HTTPS module mirror interactions.
+		See 'go help goauth'.
 	GOBIN
 		The directory where 'go install' will install a command.
 	GOCACHE
 		The directory where the go command will store cached
-		information for reuse in future builds.
-	GOMODCACHE
-		The directory where the go command will store downloaded modules.
+		information for reuse in future builds. Must be an absolute path.
+	GOCACHEPROG
+		A command (with optional space-separated flags) that implements an
+		external go command build cache.
+		See 'go doc cmd/go/internal/cacheprog'.
 	GODEBUG
-		Enable various debugging facilities. See 'go doc runtime'
-		for details.
+		Enable various debugging facilities for programs built with Go,
+		including the go command. Cannot be set using 'go env -w'.
+		See https://go.dev/doc/godebug for details.
 	GOENV
 		The location of the Go environment configuration file.
 		Cannot be set using 'go env -w'.
@@ -521,19 +552,21 @@ General-purpose environment variables:
 		manner. Only applies to dependencies that are being fetched directly.
 		GOINSECURE does not disable checksum database validation. GOPRIVATE or
 		GONOSUMDB may be used to achieve that.
+	GOMODCACHE
+		The directory where the go command will store downloaded modules.
 	GOOS
 		The operating system for which to compile code.
 		Examples are linux, darwin, windows, netbsd.
 	GOPATH
-		For more details see: 'go help gopath'.
-	GOPROXY
-		URL of Go module proxy. See https://golang.org/ref/mod#environment-variables
-		and https://golang.org/ref/mod#module-proxy for details.
+		Controls where various files are stored. See: 'go help gopath'.
 	GOPRIVATE, GONOPROXY, GONOSUMDB
 		Comma-separated list of glob patterns (in the syntax of Go's path.Match)
 		of module path prefixes that should always be fetched directly
 		or that should not be compared against the checksum database.
 		See https://golang.org/ref/mod#private-modules.
+	GOPROXY
+		URL of Go module proxy. See https://golang.org/ref/mod#environment-variables
+		and https://golang.org/ref/mod#module-proxy for details.
 	GOROOT
 		The root of the go tree.
 	GOSUMDB
@@ -542,6 +575,8 @@ General-purpose environment variables:
 	GOTMPDIR
 		The directory where the go command will write
 		temporary source files, packages, and binaries.
+	GOTOOLCHAIN
+		Controls which Go toolchain is used. See https://go.dev/doc/toolchain.
 	GOVCS
 		Lists version control commands that may be used with matching servers.
 		See 'go help vcs'.
@@ -562,8 +597,6 @@ Environment variables for use with cgo:
 		The default is 'ar'.
 	CC
 		The command to use to compile C code.
-	CGO_ENABLED
-		Whether the cgo command is supported. Either 0 or 1.
 	CGO_CFLAGS
 		Flags that cgo will pass to the compiler when compiling
 		C code.
@@ -581,6 +614,8 @@ Environment variables for use with cgo:
 	CGO_CXXFLAGS, CGO_CXXFLAGS_ALLOW, CGO_CXXFLAGS_DISALLOW
 		Like CGO_CFLAGS, CGO_CFLAGS_ALLOW, and CGO_CFLAGS_DISALLOW,
 		but for the C++ compiler.
+	CGO_ENABLED
+		Whether the cgo command is supported. Either 0 or 1.
 	CGO_FFLAGS, CGO_FFLAGS_ALLOW, CGO_FFLAGS_DISALLOW
 		Like CGO_CFLAGS, CGO_CFLAGS_ALLOW, and CGO_CFLAGS_DISALLOW,
 		but for the Fortran compiler.
@@ -596,9 +631,6 @@ Environment variables for use with cgo:
 
 Architecture-specific environment variables:
 
-	GOARM
-		For GOARCH=arm, the ARM architecture for which to compile.
-		Valid values are 5, 6, 7.
 	GO386
 		For GOARCH=386, how to implement floating point instructions.
 		Valid values are sse2 (default), softfloat.
@@ -606,6 +638,23 @@ Architecture-specific environment variables:
 		For GOARCH=amd64, the microarchitecture level for which to compile.
 		Valid values are v1 (default), v2, v3, v4.
 		See https://golang.org/wiki/MinimumRequirements#amd64
+	GOARM
+		For GOARCH=arm, the ARM architecture for which to compile.
+		Valid values are 5, 6, 7.
+		When the Go tools are built on an arm system,
+		the default value is set based on what the build system supports.
+		When the Go tools are not built on an arm system
+		(that is, when building a cross-compiler),
+		the default value is 7.
+		The value can be followed by an option specifying how to implement floating point instructions.
+		Valid options are ,softfloat (default for 5) and ,hardfloat (default for 6 and 7).
+	GOARM64
+		For GOARCH=arm64, the ARM64 architecture for which to compile.
+		Valid values are v8.0 (default), v8.{1-9}, v9.{0-5}.
+		The value can be followed by an option specifying extensions implemented by target hardware.
+		Valid options are ,lse and ,crypto.
+		Note that some extensions are enabled by default starting from a certain GOARM64 version;
+		for example, lse is enabled by default starting from v8.1.
 	GOMIPS
 		For GOARCH=mips{,le}, whether to use floating point instructions.
 		Valid values are hardfloat (default), softfloat.
@@ -615,6 +664,11 @@ Architecture-specific environment variables:
 	GOPPC64
 		For GOARCH=ppc64{,le}, the target ISA (Instruction Set Architecture).
 		Valid values are power8 (default), power9, power10.
+	GORISCV64
+		For GOARCH=riscv64, the RISC-V user-mode application profile for which
+		to compile. Valid values are rva20u64 (default), rva22u64, rva23u64.
+		See https://github.com/riscv/riscv-profiles/blob/main/src/profiles.adoc
+		and https://github.com/riscv/riscv-profiles/blob/main/src/rva23-profile.adoc
 	GOWASM
 		For GOARCH=wasm, comma-separated list of experimental WebAssembly features to use.
 		Valid values are satconv, signext.
@@ -624,7 +678,6 @@ Environment variables for use with code coverage:
 	GOCOVERDIR
 		Directory into which to write code coverage data files
 		generated by running a "go build -cover" binary.
-		Requires that GOEXPERIMENT=coverageredesign is enabled.
 
 Special-purpose environment variables:
 
@@ -634,14 +687,15 @@ Special-purpose environment variables:
 	GOEXPERIMENT
 		Comma-separated list of toolchain experiments to enable or disable.
 		The list of available experiments may change arbitrarily over time.
-		See src/internal/goexperiment/flags.go for currently valid values.
+		See GOROOT/src/internal/goexperiment/flags.go for currently valid values.
 		Warning: This variable is provided for the development and testing
 		of the Go toolchain itself. Use beyond that purpose is unsupported.
-	GOROOT_FINAL
-		The root of the installed Go tree, when it is
-		installed in a location other than where it is built.
-		File names in stack traces are rewritten from GOROOT to
-		GOROOT_FINAL.
+	GOFIPS140
+		The FIPS-140 cryptography mode to use when building binaries.
+		The default is GOFIPS140=off, which makes no FIPS-140 changes at all.
+		Other values enable FIPS-140 compliance measures and select alternate
+		versions of the cryptography source code.
+		See https://go.dev/security/fips140 for details.
 	GO_EXTLINK_ENABLED
 		Whether the linker should use external linking mode
 		when using -linkmode=auto with code that uses cgo.
@@ -668,6 +722,11 @@ Additional information available from 'go env' but not read from the environment
 		If module-aware mode is enabled, but there is no go.mod, GOMOD will be
 		os.DevNull ("/dev/null" on Unix-like systems, "NUL" on Windows).
 		If module-aware mode is disabled, GOMOD will be the empty string.
+	GOTELEMETRY
+		The current Go telemetry mode ("off", "local", or "on").
+		See "go help telemetry" for more information.
+	GOTELEMETRYDIR
+		The directory Go telemetry data is written is written to.
 	GOTOOLDIR
 		The directory where the go tools (compile, cover, doc, etc...) are installed.
 	GOVERSION
@@ -736,7 +795,10 @@ are:
 		Build the listed main package, plus all packages it imports,
 		into a C shared library. The only callable symbols will
 		be those functions exported using a cgo //export comment.
-		Requires exactly one main package to be listed.
+		On wasip1, this mode builds it to a WASI reactor/library,
+		of which the callable symbols are those functions exported
+		using a //go:wasmexport directive. Requires exactly one
+		main package to be listed.
 
 	-buildmode=default
 		Listed main packages are built into executables and listed
@@ -773,6 +835,7 @@ var HelpCache = &base.Command{
 The go command caches build outputs for reuse in future builds.
 The default location for cache data is a subdirectory named go-build
 in the standard user cache directory for the current operating system.
+The cache is safe for concurrent invocations of the go command.
 Setting the GOCACHE environment variable overrides this default,
 and running 'go env GOCACHE' prints the current cache directory.
 
@@ -825,9 +888,12 @@ line comment that begins
 
 	//go:build
 
+Build constraints can also be used to downgrade the language version
+used to compile a file.
+
 Constraints may appear in any kind of source file (not just Go), but
 they must appear near the top of the file, preceded
-only by blank lines and other line comments. These rules mean that in Go
+only by blank lines and other comments. These rules mean that in Go
 files a build constraint must appear before the package clause.
 
 To distinguish build constraints from package documentation,
@@ -890,6 +956,8 @@ The defined architecture feature build tags are:
 	  correspond to the amd64.v1, amd64.v2, and amd64.v3 feature build tags.
 	- For GOARCH=arm, GOARM=5, 6, and 7
 	  correspond to the arm.5, arm.6, and arm.7 feature build tags.
+	- For GOARCH=arm64, GOARM64=v8.{0-9} and v9.{0-5}
+	  correspond to the arm64.v8.{0-9} and arm64.v9.{0-5} feature build tags.
 	- For GOARCH=mips or mipsle,
 	  GOMIPS=hardfloat and softfloat
 	  correspond to the mips.hardfloat and mips.softfloat
@@ -903,10 +971,13 @@ The defined architecture feature build tags are:
 	  ppc64.power8, ppc64.power9, and ppc64.power10
 	  (or ppc64le.power8, ppc64le.power9, and ppc64le.power10)
 	  feature build tags.
+	- For GOARCH=riscv64,
+	  GORISCV64=rva20u64, rva22u64 and rva23u64 correspond to the riscv64.rva20u64,
+	  riscv64.rva22u64 and riscv64.rva23u64 build tags.
 	- For GOARCH=wasm, GOWASM=satconv and signext
 	  correspond to the wasm.satconv and wasm.signext feature build tags.
 
-For GOARCH=amd64, arm, ppc64, and ppc64le, a particular feature level
+For GOARCH=amd64, arm, ppc64, ppc64le, and riscv64, a particular feature level
 sets the feature build tags for all previous levels as well.
 For example, GOAMD64=v2 sets the amd64.v1 and amd64.v2 feature flags.
 This ensures that code making use of v2 features continues to compile
@@ -939,5 +1010,120 @@ only when building the package for 32-bit x86.
 Go versions 1.16 and earlier used a different syntax for build constraints,
 with a "// +build" prefix. The gofmt command will add an equivalent //go:build
 constraint when encountering the older syntax.
+
+In modules with a Go version of 1.21 or later, if a file's build constraint
+has a term for a Go major release, the language version used when compiling
+the file will be the minimum version implied by the build constraint.
 `,
 }
+
+var HelpGoAuth = &base.Command{
+	UsageLine: "goauth",
+	Short:     "GOAUTH environment variable",
+	Long: `
+GOAUTH is a semicolon-separated list of authentication commands for go-import and
+HTTPS module mirror interactions. The default is netrc.
+
+The supported authentication commands are:
+
+off
+	Disables authentication.
+netrc
+	Uses credentials from NETRC or the .netrc file in your home directory.
+git dir
+	Runs 'git credential fill' in dir and uses its credentials. The
+	go command will run 'git credential approve/reject' to update
+	the credential helper's cache.
+command
+	Executes the given command (a space-separated argument list) and attaches
+	the provided headers to HTTPS requests.
+	The command must produce output in the following format:
+		Response      = { CredentialSet } .
+		CredentialSet = URLLine { URLLine } BlankLine { HeaderLine } BlankLine .
+		URLLine       = /* URL that starts with "https://" */ '\n' .
+		HeaderLine    = /* HTTP Request header */ '\n' .
+		BlankLine     = '\n' .
+
+	Example:
+		https://example.com
+		https://example.net/api/
+
+		Authorization: Basic <token>
+
+		https://another-example.org/
+
+		Example: Data
+
+	If the server responds with any 4xx code, the go command will write the
+	following to the program's stdin:
+		Response      = StatusLine { HeaderLine } BlankLine .
+		StatusLine    = Protocol Space Status '\n' .
+		Protocol      = /* HTTP protocol */ .
+		Space         = ' ' .
+		Status        = /* HTTP status code */ .
+		BlankLine     = '\n' .
+		HeaderLine    = /* HTTP Response's header */ '\n' .
+
+	Example:
+		HTTP/1.1 401 Unauthorized
+		Content-Length: 19
+		Content-Type: text/plain; charset=utf-8
+		Date: Thu, 07 Nov 2024 18:43:09 GMT
+
+	Note: it is safe to use net/http.ReadResponse to parse this input.
+
+Before the first HTTPS fetch, the go command will invoke each GOAUTH
+command in the list with no additional arguments and no input.
+If the server responds with any 4xx code, the go command will invoke the
+GOAUTH commands again with the URL as an additional command-line argument
+and the HTTP Response to the program's stdin.
+If the server responds with an error again, the fetch fails: a URL-specific
+GOAUTH will only be attempted once per fetch.
+`,
+}
+
+var HelpBuildJSON = &base.Command{
+	UsageLine: "buildjson",
+	Short:     "build -json encoding",
+	Long: `
+The 'go build', 'go install', and 'go test' commands take a -json flag that
+reports build output and failures as structured JSON output on standard
+output.
+
+The JSON stream is a newline-separated sequence of BuildEvent objects
+corresponding to the Go struct:
+
+	type BuildEvent struct {
+		ImportPath string
+		Action     string
+		Output     string
+	}
+
+The ImportPath field gives the package ID of the package being built.
+This matches the Package.ImportPath field of go list -json and the
+TestEvent.FailedBuild field of go test -json. Note that it does not
+match TestEvent.Package.
+
+The Action field is one of the following:
+
+	build-output - The toolchain printed output
+	build-fail - The build failed
+
+The Output field is set for Action == "build-output" and is a portion of
+the build's output. The concatenation of the Output fields of all output
+events is the exact output of the build. A single event may contain one
+or more lines of output and there may be more than one output event for
+a given ImportPath. This matches the definition of the TestEvent.Output
+field produced by go test -json.
+
+For go test -json, this struct is designed so that parsers can distinguish
+interleaved TestEvents and BuildEvents by inspecting the Action field.
+Furthermore, as with TestEvent, parsers can simply concatenate the Output
+fields of all events to reconstruct the text format output, as it would
+have appeared from go build without the -json flag.
+
+Note that there may also be non-JSON error text on standard error, even
+with the -json flag. Typically, this indicates an early, serious error.
+Consumers should be robust to this.
+	`,
+}
