@@ -17,6 +17,8 @@ import (
 	"path"
 	"path/filepath"
 	"strings"
+
+	"golang.org/x/mod/modfile"
 )
 
 // A Match represents the result of matching a single package pattern.
@@ -60,14 +62,14 @@ func (m *Match) IsLocal() bool {
 }
 
 // IsMeta reports whether the pattern is a “meta-package” keyword that represents
-// multiple packages, such as "std", "cmd", "tool", or "all".
+// multiple packages, such as "std", "cmd", "tool", "work", or "all".
 func (m *Match) IsMeta() bool {
 	return IsMetaPackage(m.pattern)
 }
 
 // IsMetaPackage checks if name is a reserved package name that expands to multiple packages.
 func IsMetaPackage(name string) bool {
-	return name == "std" || name == "cmd" || name == "tool" || name == "all"
+	return name == "std" || name == "cmd" || name == "tool" || name == "work" || name == "all"
 }
 
 // A MatchError indicates an error that occurred while attempting to match a
@@ -208,6 +210,69 @@ func (m *Match) MatchPackages() {
 	}
 }
 
+// IgnorePatterns is normalized with normalizePath.
+type IgnorePatterns struct {
+	relativePatterns []string
+	anyPatterns      []string
+}
+
+// ShouldIgnore returns true if the given directory should be ignored
+// based on the ignore patterns.
+//
+// An ignore pattern "x" will cause any file or directory named "x"
+// (and its entire subtree) to be ignored, regardless of its location
+// within the module.
+//
+// An ignore pattern "./x" will only cause the specific file or directory
+// named "x" at the root of the module to be ignored.
+// Wildcards in ignore patterns are not supported.
+func (ignorePatterns *IgnorePatterns) ShouldIgnore(dir string) bool {
+	if dir == "" {
+		return false
+	}
+	dir = normalizePath(dir)
+	for _, pattern := range ignorePatterns.relativePatterns {
+		if strings.HasPrefix(dir, pattern) {
+			return true
+		}
+	}
+	for _, pattern := range ignorePatterns.anyPatterns {
+		if strings.Contains(dir, pattern) {
+			return true
+		}
+	}
+	return false
+}
+
+func NewIgnorePatterns(patterns []string) *IgnorePatterns {
+	var relativePatterns, anyPatterns []string
+	for _, pattern := range patterns {
+		ignorePatternPath, isRelative := strings.CutPrefix(pattern, "./")
+		ignorePatternPath = normalizePath(ignorePatternPath)
+		if isRelative {
+			relativePatterns = append(relativePatterns, ignorePatternPath)
+		} else {
+			anyPatterns = append(anyPatterns, ignorePatternPath)
+		}
+	}
+	return &IgnorePatterns{
+		relativePatterns: relativePatterns,
+		anyPatterns:      anyPatterns,
+	}
+}
+
+// normalizePath adds slashes to the front and end of the given path.
+func normalizePath(path string) string {
+	path = filepath.ToSlash(path)
+	if !strings.HasPrefix(path, "/") {
+		path = "/" + path
+	}
+	if !strings.HasSuffix(path, "/") {
+		path += "/"
+	}
+	return path
+}
+
 // MatchDirs sets m.Dirs to a non-nil slice containing all directories that
 // potentially match a local pattern. The pattern must begin with an absolute
 // path, or "./", or "../". On Windows, the pattern may use slash or backslash
@@ -253,16 +318,18 @@ func (m *Match) MatchDirs(modRoots []string) {
 	// We need to preserve the ./ for pattern matching
 	// and in the returned import paths.
 
-	if len(modRoots) > 1 {
+	var modRoot string
+	if len(modRoots) > 0 {
 		abs, err := filepath.Abs(dir)
 		if err != nil {
 			m.AddError(err)
 			return
 		}
 		var found bool
-		for _, modRoot := range modRoots {
-			if modRoot != "" && str.HasFilePathPrefix(abs, modRoot) {
+		for _, mr := range modRoots {
+			if mr != "" && str.HasFilePathPrefix(abs, mr) {
 				found = true
+				modRoot = mr
 			}
 		}
 		if !found {
@@ -274,6 +341,7 @@ func (m *Match) MatchDirs(modRoots []string) {
 		}
 	}
 
+	ignorePatterns := parseIgnorePatterns(modRoot)
 	// If dir is actually a symlink to a directory,
 	// we want to follow it (see https://go.dev/issue/50807).
 	// Add a trailing separator to force that to happen.
@@ -305,6 +373,17 @@ func (m *Match) MatchDirs(modRoots []string) {
 		if dot || strings.HasPrefix(elem, "_") || elem == "testdata" {
 			return filepath.SkipDir
 		}
+		absPath, err := filepath.Abs(path)
+		if err != nil {
+			return err
+		}
+
+		if ignorePatterns != nil && ignorePatterns.ShouldIgnore(InDir(absPath, modRoot)) {
+			if cfg.BuildX {
+				fmt.Fprintf(os.Stderr, "# ignoring directory %s\n", absPath)
+			}
+			return filepath.SkipDir
+		}
 
 		if !top && cfg.ModulesEnabled {
 			// Ignore other modules found in subdirectories.
@@ -353,20 +432,20 @@ func WarnUnmatched(matches []*Match) {
 
 // ImportPaths returns the matching paths to use for the given command line.
 // It calls ImportPathsQuiet and then WarnUnmatched.
-func ImportPaths(patterns, modRoots []string) []*Match {
-	matches := ImportPathsQuiet(patterns, modRoots)
+func ImportPaths(patterns []string) []*Match {
+	matches := ImportPathsQuiet(patterns)
 	WarnUnmatched(matches)
 	return matches
 }
 
 // ImportPathsQuiet is like ImportPaths but does not warn about patterns with no matches.
-func ImportPathsQuiet(patterns, modRoots []string) []*Match {
+func ImportPathsQuiet(patterns []string) []*Match {
 	patterns = CleanPatterns(patterns)
 	out := make([]*Match, 0, len(patterns))
 	for _, a := range patterns {
 		m := NewMatch(a)
 		if m.IsLocal() {
-			m.MatchDirs(modRoots)
+			m.MatchDirs(nil)
 
 			// Change the file import path to a regular import path if the package
 			// is in GOPATH or GOROOT. We don't report errors here; LoadImport
@@ -419,9 +498,7 @@ func CleanPatterns(patterns []string) []string {
 		if filepath.IsAbs(p) {
 			p = filepath.Clean(p)
 		} else {
-			if filepath.Separator == '\\' {
-				p = strings.ReplaceAll(p, `\`, `/`)
-			}
+			p = strings.ReplaceAll(p, `\`, `/`)
 
 			// Put argument in canonical form, but preserve leading ./.
 			if strings.HasPrefix(p, "./") {
@@ -511,3 +588,25 @@ func InDir(path, dir string) string {
 	}
 	return ""
 }
+
+// parseIgnorePatterns reads the go.mod file at the given module root
+// and extracts the ignore patterns defined within it.
+// If modRoot is empty, it returns nil.
+func parseIgnorePatterns(modRoot string) *IgnorePatterns {
+	if modRoot == "" {
+		return nil
+	}
+	data, err := os.ReadFile(filepath.Join(modRoot, "go.mod"))
+	if err != nil {
+		return nil
+	}
+	modFile, err := modfile.Parse("go.mod", data, nil)
+	if err != nil {
+		return nil
+	}
+	var patterns []string
+	for _, i := range modFile.Ignore {
+		patterns = append(patterns, i.Path)
+	}
+	return NewIgnorePatterns(patterns)
+}
