@@ -27,6 +27,7 @@ import (
 type gccgoToolchain struct{}
 
 var GccgoName, GccgoBin string
+var GccgoChanged bool
 var gccgoErr error
 
 func init() {
@@ -34,6 +35,9 @@ func init() {
 	if GccgoName == "" {
 		GccgoName = "gccgo"
 	}
+	if GccgoName != "gccgo" {
+		GccgoChanged = true
+	}
 	GccgoBin, gccgoErr = pathcache.LookPath(GccgoName)
 }
 
