@@ -317,6 +317,11 @@ const (
 	// it is the first instruction in an AUIPC + S-type pair that needs a
 	// R_RISCV_PCREL_STYPE relocation.
 	NEED_PCREL_STYPE_RELOC
+
+	// NEED_GOT_PCREL_ITYPE_RELOC is set on AUIPC instructions to indicate that
+	// it is the first instruction in an AUIPC + I-type pair that needs a
+	// R_RISCV_GOT_PCREL_ITYPE relocation.
+	NEED_GOT_PCREL_ITYPE_RELOC
 )
 
 // RISC-V mnemonics, as defined in the "opcodes" and "opcodes-pseudo" files
@@ -571,6 +576,10 @@ const (
 	// 22.5 Quad-Precision Floating-Point Classify Instruction
 	AFCLASSQ
 
+	//
+	// "B" Extension for Bit Manipulation, Version 1.0.0
+	//
+
 	// 28.4.1: Address Generation Instructions (Zba)
 	AADDUW
 	ASH1ADD
@@ -620,15 +629,15 @@ const (
 	ABSETI
 
 	//
-	// RISC-V Vector ISA-extension (1.0) (Unprivileged 20240411)
+	// "V" Standard Extension for Vector Operations, Version 1.0
 	//
 
-	// 31.6. Configuration-Setting Instructions
+	// 31.6: Configuration-Setting Instructions
 	AVSETVLI
 	AVSETIVLI
 	AVSETVL
 
-	// 31.7.4. Vector Unit-Stride Instructions
+	// 31.7.4: Vector Unit-Stride Instructions
 	AVLE8V
 	AVLE16V
 	AVLE32V
@@ -640,7 +649,7 @@ const (
 	AVLMV
 	AVSMV
 
-	// 31.7.5. Vector Strided Instructions
+	// 31.7.5: Vector Strided Instructions
 	AVLSE8V
 	AVLSE16V
 	AVLSE32V
@@ -650,7 +659,7 @@ const (
 	AVSSE32V
 	AVSSE64V
 
-	// 31.7.6. Vector Indexed Instructions
+	// 31.7.6: Vector Indexed Instructions
 	AVLUXEI8V
 	AVLUXEI16V
 	AVLUXEI32V
@@ -668,13 +677,13 @@ const (
 	AVSOXEI32V
 	AVSOXEI64V
 
-	// 31.7.7. Unit-stride Fault-Only-First Loads
+	// 31.7.7: Unit-stride Fault-Only-First Loads
 	AVLE8FFV
 	AVLE16FFV
 	AVLE32FFV
 	AVLE64FFV
 
-	// 31.7.9. Vector Load/Store Whole Register Instructions
+	// 31.7.9: Vector Load/Store Whole Register Instructions
 	AVL1RE8V
 	AVL1RE16V
 	AVL1RE32V
@@ -696,7 +705,7 @@ const (
 	AVS4RV
 	AVS8RV
 
-	// 31.11.1. Vector Single-Width Integer Add and Subtract
+	// 31.11.1: Vector Single-Width Integer Add and Subtract
 	AVADDVV
 	AVADDVX
 	AVADDVI
@@ -705,7 +714,7 @@ const (
 	AVRSUBVX
 	AVRSUBVI
 
-	// 31.11.2. Vector Widening Integer Add/Subtract
+	// 31.11.2: Vector Widening Integer Add/Subtract
 	AVWADDUVV
 	AVWADDUVX
 	AVWSUBUVV
@@ -723,7 +732,7 @@ const (
 	AVWSUBWV
 	AVWSUBWX
 
-	// 31.11.3. Vector Integer Extension
+	// 31.11.3: Vector Integer Extension
 	AVZEXTVF2
 	AVSEXTVF2
 	AVZEXTVF4
@@ -731,7 +740,7 @@ const (
 	AVZEXTVF8
 	AVSEXTVF8
 
-	// 31.11.4. Vector Integer Add-with-Carry / Subtract-with-Borrow Instructions
+	// 31.11.4: Vector Integer Add-with-Carry / Subtract-with-Borrow Instructions
 	AVADCVVM
 	AVADCVXM
 	AVADCVIM
@@ -748,7 +757,7 @@ const (
 	AVMSBCVV
 	AVMSBCVX
 
-	// 31.11.5. Vector Bitwise Logical Instructions
+	// 31.11.5: Vector Bitwise Logical Instructions
 	AVANDVV
 	AVANDVX
 	AVANDVI
@@ -759,7 +768,7 @@ const (
 	AVXORVX
 	AVXORVI
 
-	// 31.11.6. Vector Single-Width Shift Instructions
+	// 31.11.6: Vector Single-Width Shift Instructions
 	AVSLLVV
 	AVSLLVX
 	AVSLLVI
@@ -770,7 +779,7 @@ const (
 	AVSRAVX
 	AVSRAVI
 
-	// 31.11.7. Vector Narrowing Integer Right Shift Instructions
+	// 31.11.7: Vector Narrowing Integer Right Shift Instructions
 	AVNSRLWV
 	AVNSRLWX
 	AVNSRLWI
@@ -778,7 +787,7 @@ const (
 	AVNSRAWX
 	AVNSRAWI
 
-	// 31.11.8. Vector Integer Compare Instructions
+	// 31.11.8: Vector Integer Compare Instructions
 	AVMSEQVV
 	AVMSEQVX
 	AVMSEQVI
@@ -800,7 +809,7 @@ const (
 	AVMSGTVX
 	AVMSGTVI
 
-	// 31.11.9. Vector Integer Min/Max Instructions
+	// 31.11.9: Vector Integer Min/Max Instructions
 	AVMINUVV
 	AVMINUVX
 	AVMINVV
@@ -810,7 +819,7 @@ const (
 	AVMAXVV
 	AVMAXVX
 
-	// 31.11.10. Vector Single-Width Integer Multiply Instructions
+	// 31.11.10: Vector Single-Width Integer Multiply Instructions
 	AVMULVV
 	AVMULVX
 	AVMULHVV
@@ -820,7 +829,7 @@ const (
 	AVMULHSUVV
 	AVMULHSUVX
 
-	// 31.11.11. Vector Integer Divide Instructions
+	// 31.11.11: Vector Integer Divide Instructions
 	AVDIVUVV
 	AVDIVUVX
 	AVDIVVV
@@ -830,7 +839,7 @@ const (
 	AVREMVV
 	AVREMVX
 
-	// 31.11.12. Vector Widening Integer Multiply Instructions
+	// 31.11.12: Vector Widening Integer Multiply Instructions
 	AVWMULVV
 	AVWMULVX
 	AVWMULUVV
@@ -838,7 +847,7 @@ const (
 	AVWMULSUVV
 	AVWMULSUVX
 
-	// 31.11.13. Vector Single-Width Integer Multiply-Add Instructions
+	// 31.11.13: Vector Single-Width Integer Multiply-Add Instructions
 	AVMACCVV
 	AVMACCVX
 	AVNMSACVV
@@ -848,7 +857,7 @@ const (
 	AVNMSUBVV
 	AVNMSUBVX
 
-	// 31.11.14. Vector Widening Integer Multiply-Add Instructions
+	// 31.11.14: Vector Widening Integer Multiply-Add Instructions
 	AVWMACCUVV
 	AVWMACCUVX
 	AVWMACCVV
@@ -857,17 +866,17 @@ const (
 	AVWMACCSUVX
 	AVWMACCUSVX
 
-	// 31.11.15. Vector Integer Merge Instructions
+	// 31.11.15: Vector Integer Merge Instructions
 	AVMERGEVVM
 	AVMERGEVXM
 	AVMERGEVIM
 
-	// 31.11.16. Vector Integer Move Instructions
+	// 31.11.16: Vector Integer Move Instructions
 	AVMVVV
 	AVMVVX
 	AVMVVI
 
-	// 31.12.1. Vector Single-Width Saturating Add and Subtract
+	// 31.12.1: Vector Single-Width Saturating Add and Subtract
 	AVSADDUVV
 	AVSADDUVX
 	AVSADDUVI
@@ -879,7 +888,7 @@ const (
 	AVSSUBVV
 	AVSSUBVX
 
-	// 31.12.2. Vector Single-Width Averaging Add and Subtract
+	// 31.12.2: Vector Single-Width Averaging Add and Subtract
 	AVAADDUVV
 	AVAADDUVX
 	AVAADDVV
@@ -889,11 +898,11 @@ const (
 	AVASUBVV
 	AVASUBVX
 
-	// 31.12.3. Vector Single-Width Fractional Multiply with Rounding and Saturation
+	// 31.12.3: Vector Single-Width Fractional Multiply with Rounding and Saturation
 	AVSMULVV
 	AVSMULVX
 
-	// 31.12.4. Vector Single-Width Scaling Shift Instructions
+	// 31.12.4: Vector Single-Width Scaling Shift Instructions
 	AVSSRLVV
 	AVSSRLVX
 	AVSSRLVI
@@ -901,7 +910,7 @@ const (
 	AVSSRAVX
 	AVSSRAVI
 
-	// 31.12.5. Vector Narrowing Fixed-Point Clip Instructions
+	// 31.12.5: Vector Narrowing Fixed-Point Clip Instructions
 	AVNCLIPUWV
 	AVNCLIPUWX
 	AVNCLIPUWI
@@ -909,14 +918,14 @@ const (
 	AVNCLIPWX
 	AVNCLIPWI
 
-	// 31.13.2. Vector Single-Width Floating-Point Add/Subtract Instructions
+	// 31.13.2: Vector Single-Width Floating-Point Add/Subtract Instructions
 	AVFADDVV
 	AVFADDVF
 	AVFSUBVV
 	AVFSUBVF
 	AVFRSUBVF
 
-	// 31.13.3. Vector Widening Floating-Point Add/Subtract Instructions
+	// 31.13.3: Vector Widening Floating-Point Add/Subtract Instructions
 	AVFWADDVV
 	AVFWADDVF
 	AVFWSUBVV
@@ -926,18 +935,18 @@ const (
 	AVFWSUBWV
 	AVFWSUBWF
 
-	// 31.13.4. Vector Single-Width Floating-Point Multiply/Divide Instructions
+	// 31.13.4: Vector Single-Width Floating-Point Multiply/Divide Instructions
 	AVFMULVV
 	AVFMULVF
 	AVFDIVVV
 	AVFDIVVF
 	AVFRDIVVF
 
-	// 31.13.5. Vector Widening Floating-Point Multiply
+	// 31.13.5: Vector Widening Floating-Point Multiply
 	AVFWMULVV
 	AVFWMULVF
 
-	// 31.13.6. Vector Single-Width Floating-Point Fused Multiply-Add Instructions
+	// 31.13.6: Vector Single-Width Floating-Point Fused Multiply-Add Instructions
 	AVFMACCVV
 	AVFMACCVF
 	AVFNMACCVV
@@ -955,7 +964,7 @@ const (
 	AVFNMSUBVV
 	AVFNMSUBVF
 
-	// 31.13.7. Vector Widening Floating-Point Fused Multiply-Add Instructions
+	// 31.13.7: Vector Widening Floating-Point Fused Multiply-Add Instructions
 	AVFWMACCVV
 	AVFWMACCVF
 	AVFWNMACCVV
@@ -965,22 +974,22 @@ const (
 	AVFWNMSACVV
 	AVFWNMSACVF
 
-	// 31.13.8. Vector Floating-Point Square-Root Instruction
+	// 31.13.8: Vector Floating-Point Square-Root Instruction
 	AVFSQRTV
 
-	// 31.13.9. Vector Floating-Point Reciprocal Square-Root Estimate Instruction
+	// 31.13.9: Vector Floating-Point Reciprocal Square-Root Estimate Instruction
 	AVFRSQRT7V
 
-	// 31.13.10. Vector Floating-Point Reciprocal Estimate Instruction
+	// 31.13.10: Vector Floating-Point Reciprocal Estimate Instruction
 	AVFREC7V
 
-	// 31.13.11. Vector Floating-Point MIN/MAX Instructions
+	// 31.13.11: Vector Floating-Point MIN/MAX Instructions
 	AVFMINVV
 	AVFMINVF
 	AVFMAXVV
 	AVFMAXVF
 
-	// 31.13.12. Vector Floating-Point Sign-Injection Instructions
+	// 31.13.12: Vector Floating-Point Sign-Injection Instructions
 	AVFSGNJVV
 	AVFSGNJVF
 	AVFSGNJNVV
@@ -988,7 +997,7 @@ const (
 	AVFSGNJXVV
 	AVFSGNJXVF
 
-	// 31.13.13. Vector Floating-Point Compare Instructions
+	// 31.13.13: Vector Floating-Point Compare Instructions
 	AVMFEQVV
 	AVMFEQVF
 	AVMFNEVV
@@ -1000,16 +1009,16 @@ const (
 	AVMFGTVF
 	AVMFGEVF
 
-	// 31.13.14. Vector Floating-Point Classify Instruction
+	// 31.13.14: Vector Floating-Point Classify Instruction
 	AVFCLASSV
 
-	// 31.13.15. Vector Floating-Point Merge Instruction
+	// 31.13.15: Vector Floating-Point Merge Instruction
 	AVFMERGEVFM
 
-	// 31.13.16. Vector Floating-Point Move Instruction
+	// 31.13.16: Vector Floating-Point Move Instruction
 	AVFMVVF
 
-	// 31.13.17. Single-Width Floating-Point/Integer Type-Convert Instructions
+	// 31.13.17: Single-Width Floating-Point/Integer Type-Convert Instructions
 	AVFCVTXUFV
 	AVFCVTXFV
 	AVFCVTRTZXUFV
@@ -1017,7 +1026,7 @@ const (
 	AVFCVTFXUV
 	AVFCVTFXV
 
-	// 31.13.18. Widening Floating-Point/Integer Type-Convert Instructions
+	// 31.13.18: Widening Floating-Point/Integer Type-Convert Instructions
 	AVFWCVTXUFV
 	AVFWCVTXFV
 	AVFWCVTRTZXUFV
@@ -1026,7 +1035,7 @@ const (
 	AVFWCVTFXV
 	AVFWCVTFFV
 
-	// 31.13.19. Narrowing Floating-Point/Integer Type-Convert Instructions
+	// 31.13.19: Narrowing Floating-Point/Integer Type-Convert Instructions
 	AVFNCVTXUFW
 	AVFNCVTXFW
 	AVFNCVTRTZXUFW
@@ -1036,7 +1045,7 @@ const (
 	AVFNCVTFFW
 	AVFNCVTRODFFW
 
-	// 31.14.1. Vector Single-Width Integer Reduction Instructions
+	// 31.14.1: Vector Single-Width Integer Reduction Instructions
 	AVREDSUMVS
 	AVREDMAXUVS
 	AVREDMAXVS
@@ -1046,21 +1055,21 @@ const (
 	AVREDORVS
 	AVREDXORVS
 
-	// 31.14.2. Vector Widening Integer Reduction Instructions
+	// 31.14.2: Vector Widening Integer Reduction Instructions
 	AVWREDSUMUVS
 	AVWREDSUMVS
 
-	// 31.14.3. Vector Single-Width Floating-Point Reduction Instructions
+	// 31.14.3: Vector Single-Width Floating-Point Reduction Instructions
 	AVFREDOSUMVS
 	AVFREDUSUMVS
 	AVFREDMAXVS
 	AVFREDMINVS
 
-	// 31.14.4. Vector Widening Floating-Point Reduction Instructions
+	// 31.14.4: Vector Widening Floating-Point Reduction Instructions
 	AVFWREDOSUMVS
 	AVFWREDUSUMVS
 
-	// 31.15. Vector Mask Instructions
+	// 31.15: Vector Mask Instructions
 	AVMANDMM
 	AVMNANDMM
 	AVMANDNMM
@@ -1077,15 +1086,15 @@ const (
 	AVIOTAM
 	AVIDV
 
-	// 31.16.1. Integer Scalar Move Instructions
+	// 31.16.1: Integer Scalar Move Instructions
 	AVMVXS
 	AVMVSX
 
-	// 31.16.2. Floating-Point Scalar Move Instructions
+	// 31.16.2: Floating-Point Scalar Move Instructions
 	AVFMVFS
 	AVFMVSF
 
-	// 31.16.3. Vector Slide Instructions
+	// 31.16.3: Vector Slide Instructions
 	AVSLIDEUPVX
 	AVSLIDEUPVI
 	AVSLIDEDOWNVX
@@ -1095,16 +1104,16 @@ const (
 	AVSLIDE1DOWNVX
 	AVFSLIDE1DOWNVF
 
-	// 31.16.4. Vector Register Gather Instructions
+	// 31.16.4: Vector Register Gather Instructions
 	AVRGATHERVV
 	AVRGATHEREI16VV
 	AVRGATHERVX
 	AVRGATHERVI
 
-	// 31.16.5. Vector Compress Instruction
+	// 31.16.5: Vector Compress Instruction
 	AVCOMPRESSVM
 
-	// 31.16.6. Whole Vector Register Move
+	// 31.16.6: Whole Vector Register Move
 	AVMV1RV
 	AVMV2RV
 	AVMV4RV
@@ -1169,6 +1178,31 @@ const (
 	ARDTIME
 	ASEQZ
 	ASNEZ
+	AVFABSV
+	AVFNEGV
+	AVL1RV
+	AVL2RV
+	AVL4RV
+	AVL8RV
+	AVMCLRM
+	AVMFGEVV
+	AVMFGTVV
+	AVMMVM
+	AVMNOTM
+	AVMSETM
+	AVMSGEUVI
+	AVMSGEUVV
+	AVMSGEVI
+	AVMSGEVV
+	AVMSGTUVV
+	AVMSGTVV
+	AVMSLTUVI
+	AVMSLTVI
+	AVNCVTXXW
+	AVNEGV
+	AVNOTV
+	AVWCVTUXXV
+	AVWCVTXXV
 
 	// End marker
 	ALAST
@@ -1218,6 +1252,77 @@ const (
 	RM_RMM              // Round to Nearest, ties to Max Magnitude
 )
 
+type SpecialOperand int
+
+const (
+	SPOP_BEGIN SpecialOperand = obj.SpecialOperandRISCVBase
+
+	// Vector mask policy.
+	SPOP_MA SpecialOperand = obj.SpecialOperandRISCVBase + iota - 1
+	SPOP_MU
+
+	// Vector tail policy.
+	SPOP_TA
+	SPOP_TU
+
+	// Vector register group multiplier (VLMUL).
+	SPOP_M1
+	SPOP_M2
+	SPOP_M4
+	SPOP_M8
+	SPOP_MF2
+	SPOP_MF4
+	SPOP_MF8
+
+	// Vector selected element width (VSEW).
+	SPOP_E8
+	SPOP_E16
+	SPOP_E32
+	SPOP_E64
+
+	SPOP_END
+)
+
+var specialOperands = map[SpecialOperand]struct {
+	encoding uint32
+	name     string
+}{
+	SPOP_MA: {encoding: 1, name: "MA"},
+	SPOP_MU: {encoding: 0, name: "MU"},
+
+	SPOP_TA: {encoding: 1, name: "TA"},
+	SPOP_TU: {encoding: 0, name: "TU"},
+
+	SPOP_M1:  {encoding: 0, name: "M1"},
+	SPOP_M2:  {encoding: 1, name: "M2"},
+	SPOP_M4:  {encoding: 2, name: "M4"},
+	SPOP_M8:  {encoding: 3, name: "M8"},
+	SPOP_MF8: {encoding: 5, name: "MF8"},
+	SPOP_MF4: {encoding: 6, name: "MF4"},
+	SPOP_MF2: {encoding: 7, name: "MF2"},
+
+	SPOP_E8:  {encoding: 0, name: "E8"},
+	SPOP_E16: {encoding: 1, name: "E16"},
+	SPOP_E32: {encoding: 2, name: "E32"},
+	SPOP_E64: {encoding: 3, name: "E64"},
+}
+
+func (so SpecialOperand) encode() uint32 {
+	op, ok := specialOperands[so]
+	if ok {
+		return op.encoding
+	}
+	return 0
+}
+
+func (so SpecialOperand) String() string {
+	op, ok := specialOperands[so]
+	if ok {
+		return op.name
+	}
+	return ""
+}
+
 // All unary instructions which write to their arguments (as opposed to reading
 // from them) go here. The assembly parser uses this information to populate
 // its AST in a semantically reasonable way.
