@@ -2,7 +2,7 @@
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
-//go:build (cgo || darwin) && !netgo && unix
+//go:build !netgo && ((cgo && unix) || darwin)
 
 package net
 
@@ -14,10 +14,7 @@ import (
 func TestCgoLookupIP(t *testing.T) {
 	defer dnsWaitGroup.Wait()
 	ctx := context.Background()
-	_, err, ok := cgoLookupIP(ctx, "ip", "localhost")
-	if !ok {
-		t.Errorf("cgoLookupIP must not be a placeholder")
-	}
+	_, err := cgoLookupIP(ctx, "ip", "localhost")
 	if err != nil {
 		t.Error(err)
 	}
@@ -27,10 +24,7 @@ func TestCgoLookupIPWithCancel(t *testing.T) {
 	defer dnsWaitGroup.Wait()
 	ctx, cancel := context.WithCancel(context.Background())
 	defer cancel()
-	_, err, ok := cgoLookupIP(ctx, "ip", "localhost")
-	if !ok {
-		t.Errorf("cgoLookupIP must not be a placeholder")
-	}
+	_, err := cgoLookupIP(ctx, "ip", "localhost")
 	if err != nil {
 		t.Error(err)
 	}
@@ -39,10 +33,7 @@ func TestCgoLookupIPWithCancel(t *testing.T) {
 func TestCgoLookupPort(t *testing.T) {
 	defer dnsWaitGroup.Wait()
 	ctx := context.Background()
-	_, err, ok := cgoLookupPort(ctx, "tcp", "smtp")
-	if !ok {
-		t.Errorf("cgoLookupPort must not be a placeholder")
-	}
+	_, err := cgoLookupPort(ctx, "tcp", "smtp")
 	if err != nil {
 		t.Error(err)
 	}
@@ -52,10 +43,7 @@ func TestCgoLookupPortWithCancel(t *testing.T) {
 	defer dnsWaitGroup.Wait()
 	ctx, cancel := context.WithCancel(context.Background())
 	defer cancel()
-	_, err, ok := cgoLookupPort(ctx, "tcp", "smtp")
-	if !ok {
-		t.Errorf("cgoLookupPort must not be a placeholder")
-	}
+	_, err := cgoLookupPort(ctx, "tcp", "smtp")
 	if err != nil {
 		t.Error(err)
 	}
@@ -64,10 +52,7 @@ func TestCgoLookupPortWithCancel(t *testing.T) {
 func TestCgoLookupPTR(t *testing.T) {
 	defer dnsWaitGroup.Wait()
 	ctx := context.Background()
-	_, err, ok := cgoLookupPTR(ctx, "127.0.0.1")
-	if !ok {
-		t.Errorf("cgoLookupPTR must not be a placeholder")
-	}
+	_, err := cgoLookupPTR(ctx, "127.0.0.1")
 	if err != nil {
 		t.Error(err)
 	}
@@ -77,10 +62,7 @@ func TestCgoLookupPTRWithCancel(t *testing.T) {
 	defer dnsWaitGroup.Wait()
 	ctx, cancel := context.WithCancel(context.Background())
 	defer cancel()
-	_, err, ok := cgoLookupPTR(ctx, "127.0.0.1")
-	if !ok {
-		t.Errorf("cgoLookupPTR must not be a placeholder")
-	}
+	_, err := cgoLookupPTR(ctx, "127.0.0.1")
 	if err != nil {
 		t.Error(err)
 	}
