@@ -4,6 +4,13 @@
 
 //go:build unix
 
+// When cross-compiling with clang to linux/armv5, atomics are emulated
+// and cause a compiler warning. This results in a build failure since
+// cgo uses -Werror. See #65290.
+#pragma GCC diagnostic ignored "-Wpragmas"
+#pragma GCC diagnostic ignored "-Wunknown-warning-option"
+#pragma GCC diagnostic ignored "-Watomic-alignment"
+
 #include <pthread.h>
 #include <errno.h>
 #include <stdio.h>
@@ -30,8 +37,13 @@ static void (*cgo_context_function)(struct context_arg*);
 
 void
 x_cgo_sys_thread_create(void* (*func)(void*), void* arg) {
+	pthread_attr_t attr;
 	pthread_t p;
-	int err = _cgo_try_pthread_create(&p, NULL, func, arg);
+	int err;
+
+	pthread_attr_init(&attr);
+	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
+	err = _cgo_try_pthread_create(&p, &attr, func, arg);
 	if (err != 0) {
 		fprintf(stderr, "pthread_create failed: %s", strerror(err));
 		abort();
@@ -41,30 +53,39 @@ x_cgo_sys_thread_create(void* (*func)(void*), void* arg) {
 uintptr_t
 _cgo_wait_runtime_init_done(void) {
 	void (*pfn)(struct context_arg*);
+	int done;
 
-	pthread_mutex_lock(&runtime_init_mu);
-	while (runtime_init_done == 0) {
-		pthread_cond_wait(&runtime_init_cond, &runtime_init_mu);
+	pfn = __atomic_load_n(&cgo_context_function, __ATOMIC_CONSUME);
+
+	done = 2;
+	if (__atomic_load_n(&runtime_init_done, __ATOMIC_CONSUME) != done) {
+		pthread_mutex_lock(&runtime_init_mu);
+		while (__atomic_load_n(&runtime_init_done, __ATOMIC_CONSUME) == 0) {
+			pthread_cond_wait(&runtime_init_cond, &runtime_init_mu);
+		}
+
+		// The key and x_cgo_pthread_key_created are for the whole program,
+		// whereas the specific and destructor is per thread.
+		if (x_cgo_pthread_key_created == 0 && pthread_key_create(&pthread_g, pthread_key_destructor) == 0) {
+			x_cgo_pthread_key_created = 1;
+		}
+
+
+		// TODO(iant): For the case of a new C thread calling into Go, such
+		// as when using -buildmode=c-archive, we know that Go runtime
+		// initialization is complete but we do not know that all Go init
+		// functions have been run. We should not fetch cgo_context_function
+		// until they have been, because that is where a call to
+		// SetCgoTraceback is likely to occur. We are going to wait for Go
+		// initialization to be complete anyhow, later, by waiting for
+		// main_init_done to be closed in cgocallbackg1. We should wait here
+		// instead. See also issue #15943.
+		pfn = __atomic_load_n(&cgo_context_function, __ATOMIC_CONSUME);
+
+		__atomic_store_n(&runtime_init_done, done, __ATOMIC_RELEASE);
+		pthread_mutex_unlock(&runtime_init_mu);
 	}
 
-	// The key and x_cgo_pthread_key_created are for the whole program,
-	// whereas the specific and destructor is per thread.
-	if (x_cgo_pthread_key_created == 0 && pthread_key_create(&pthread_g, pthread_key_destructor) == 0) {
-		x_cgo_pthread_key_created = 1;
-	}
-
-	// TODO(iant): For the case of a new C thread calling into Go, such
-	// as when using -buildmode=c-archive, we know that Go runtime
-	// initialization is complete but we do not know that all Go init
-	// functions have been run. We should not fetch cgo_context_function
-	// until they have been, because that is where a call to
-	// SetCgoTraceback is likely to occur. We are going to wait for Go
-	// initialization to be complete anyhow, later, by waiting for
-	// main_init_done to be closed in cgocallbackg1. We should wait here
-	// instead. See also issue #15943.
-	pfn = cgo_context_function;
-
-	pthread_mutex_unlock(&runtime_init_mu);
 	if (pfn != nil) {
 		struct context_arg arg;
 
@@ -75,6 +96,30 @@ _cgo_wait_runtime_init_done(void) {
 	return 0;
 }
 
+// _cgo_set_stacklo sets g->stacklo based on the stack size.
+// This is common code called from x_cgo_init, which is itself
+// called by rt0_go in the runtime package.
+void _cgo_set_stacklo(G *g, uintptr *pbounds)
+{
+	uintptr bounds[2];
+
+	// pbounds can be passed in by the caller; see gcc_linux_amd64.c.
+	if (pbounds == NULL) {
+		pbounds = &bounds[0];
+	}
+
+	x_cgo_getstackbound(pbounds);
+
+	g->stacklo = *pbounds;
+
+	// Sanity check the results now, rather than getting a
+	// morestack on g0 crash.
+	if (g->stacklo >= g->stackhi) {
+		fprintf(stderr, "runtime/cgo: bad stack bounds: lo=%p hi=%p\n", (void*)(g->stacklo), (void*)(g->stackhi));
+		abort();
+	}
+}
+
 // Store the g into a thread-specific value associated with the pthread key pthread_g.
 // And pthread_key_destructor will dropm when the thread is exiting.
 void x_cgo_bindm(void* g) {
@@ -88,7 +133,7 @@ void x_cgo_bindm(void* g) {
 void
 x_cgo_notify_runtime_init_done(void* dummy __attribute__ ((unused))) {
 	pthread_mutex_lock(&runtime_init_mu);
-	runtime_init_done = 1;
+	__atomic_store_n(&runtime_init_done, 1, __ATOMIC_RELEASE);
 	pthread_cond_broadcast(&runtime_init_cond);
 	pthread_mutex_unlock(&runtime_init_mu);
 }
@@ -96,19 +141,12 @@ x_cgo_notify_runtime_init_done(void* dummy __attribute__ ((unused))) {
 // Sets the context function to call to record the traceback context
 // when calling a Go function from C code. Called from runtime.SetCgoTraceback.
 void x_cgo_set_context_function(void (*context)(struct context_arg*)) {
-	pthread_mutex_lock(&runtime_init_mu);
-	cgo_context_function = context;
-	pthread_mutex_unlock(&runtime_init_mu);
+	__atomic_store_n(&cgo_context_function, context, __ATOMIC_RELEASE);
 }
 
 // Gets the context function.
 void (*(_cgo_get_context_function(void)))(struct context_arg*) {
-	void (*ret)(struct context_arg*);
-
-	pthread_mutex_lock(&runtime_init_mu);
-	ret = cgo_context_function;
-	pthread_mutex_unlock(&runtime_init_mu);
-	return ret;
+	return __atomic_load_n(&cgo_context_function, __ATOMIC_CONSUME);
 }
 
 // _cgo_try_pthread_create retries pthread_create if it fails with
@@ -122,7 +160,6 @@ _cgo_try_pthread_create(pthread_t* thread, const pthread_attr_t* attr, void* (*p
 	for (tries = 0; tries < 20; tries++) {
 		err = pthread_create(thread, attr, pfn, arg);
 		if (err == 0) {
-			pthread_detach(*thread);
 			return 0;
 		}
 		if (err != EAGAIN) {
